"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.compute.reducer

Implements a reducer pattern - reducers halve the size of a texture by
taking a "sampler" source which takes 4 samples from a 2x2 square of
the original texture and returns a single sample
"""

from pygl3display.infrastructure import GL3Client
from pygl3display.uniforms import UniformStore, ProxyUniform
from pygl3display.gl import GL_TRIANGLES
from pygl3display.twodshaders import TwoDShaderCommon
from pygl3display.rendertargets import TwoDTextureTarget
from pygl3display.glcommoninst import GLCommon
from pygl3display.compute.dumbtriangle import Process, makeDumbTriangles

class ReducerTriangleRenderer(GL3Client):
    """A client for Triangle renderer"""
    
    def __init__(self, shader, common):
        """Initialise the Triangle renderer.
        See GL3CLient for missing argument info
        """
        super(ReducerTriangleRenderer, self).__init__(GL_TRIANGLES, shader,
                            vertexArraySpec=common.spec)
        self.shader = shader
        self.textuProxy = ProxyUniform()
        self.textxProxy = ProxyUniform()
        self.texszProxy = ProxyUniform()
        self.uniforms = UniformStore(set([
            TwoDShaderCommon.getProjMatrix(),
            self.textuProxy, self.textxProxy, self.texszProxy,
            ]))

    def setIntex(self, newtex):
        newtex.bindTexture()
        self.textuProxy.proxy(newtex.tuUniform)
        self.textxProxy.proxy(newtex.texelUniform)
        self.texszProxy.proxy(newtex.sizeUniform)
        self.textuProxy.assignName('intex', self.shader)
        self.textxProxy.assignName('texelSize', self.shader)
        self.texszProxy.assignName('texSize', self.shader)

class ReducerProcess(Process):

    def reduce(self, texture):
        # Assumes a power of 2 texture
        size = texture.size
        while size > 1:
            size >>= 1
            self.renderer.setIntex(texture)
            target = TwoDTextureTarget(size, compute = True)
            reducers = self.triangleClass(((0, 0), (0, size), (size, 0), 
                                    (size, 0), (size, size), (0, size)), 
                                    target = target)
            GLCommon.doTasks(target)
            texture = target.texture
        return texture.readPixels()

sampleSrc = """
vec4 sample(in vec4 x0, in vec4 x1, in vec4 x2, in vec4 x3);
vec4 sample(in vec4 x0, in vec4 x1, in vec4 x2, in vec4 x3){
    vec4 ret = vec4(x0.xy, 0.0, 1.0);
    if (x1.x > ret.x) {ret.x = x1.x;}
    if (x1.y < ret.y) {ret.y = x1.y;}
    if (x2.x > ret.x) {ret.x = x2.x;}
    if (x2.y < ret.y) {ret.y = x2.y;}
    if (x3.x > ret.x) {ret.x = x3.x;}
    if (x3.y < ret.y) {ret.y = x3.y;}
    ret.z = x0.z + x1.z + x2.z + x3.z;
    return ret;
}
"""

fragmentProg = """
uniform sampler2D intex;
varying vec2 texpos;
uniform float texelSize;
uniform float texSize;
vec4 sample(in vec4 x0, in vec4 x1, in vec4 x2, in vec4 x3);

void main(){
    vec2 c0 = texpos; 
    vec2 c1 = c0;
    c1.x += texelSize;
    vec2 c2 = c0;
    c2.y += texelSize;
    vec2 c3 = c0 + texelSize;

    vec4 x0 = texture2D(intex, c0);
    vec4 x1 = texture2D(intex, c1);
    vec4 x2 = texture2D(intex, c2);
    vec4 x3 = texture2D(intex, c3);
    
    gl_FragColor = sample(x0, x1, x2, x3);
}"""


def makeReducer(apiname, sampleSrc, retLen):
    reducer = makeDumbTriangles(apiname, [sampleSrc, fragmentProg], ReducerTriangleRenderer, ReducerProcess)
    class Reducer(object):
        def __init__(self, reducer):
            self.reducer = reducer
        def reduce(self, tex):
            result = self.reducer.reduce(tex)
            return [x for x in result][:retLen]
        __call__ = reduce
    Reducer.__name__ = apiname
    return Reducer(reducer)