"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in physical 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.compute.twodsimplexnoise

An implementation of Perlins 2d Simplex Noise algorithm within a GPU shader.
"""

from __future__ import division

import random, math

from pygl3display.infrastructure import GL3Client
from pygl3display.uniforms import Uniform, UniformStore
from pygl3display.gl import glUniform1f, glUniform1i, GL_TRIANGLES
from pygl3display.onedtexture import OneDTexture
from pygl3display.twodshaders import TwoDShaderCommon
from pygl3display.rendertargets import TwoDTextureTarget
from pygl3display.glcommoninst import GLCommon
from pygl3display.compute.dumbtriangle import Process, makeDumbTriangles

class SimplexRenderer(GL3Client):
    """A client for Triangle renderer"""
    
    def __init__(self, shader, common):
        """Initialise the Triangle renderer.
        See GL3CLient for missing argument info
        """
        seed = 0
        
        random.seed(seed)
        
        self.tex = OneDTexture(256)
        self.randomise()
        super(SimplexRenderer, self).__init__(GL_TRIANGLES, shader,
                            vertexArraySpec=common.spec)
        self.freqbase = Uniform(1.0, glUniform1f, [], defaultName='freqBase')
        self.ampbase = Uniform(1.0, glUniform1f, [], defaultName='ampBase')
        self.gain = Uniform(1.0, glUniform1f, [], defaultName='gain')
        self.lacunarity = Uniform(1.0, glUniform1f, [], defaultName='lacunarity')
        self.octaves = Uniform(1.0, glUniform1i, [], defaultName='octaves')
        self.base = Uniform(1.0, glUniform1f, [], defaultName='base')
        self.mult = Uniform(1.0, glUniform1f, [], defaultName='mult')
        self.uniforms = UniformStore(set([
            TwoDShaderCommon.getProjMatrix(),
            self.freqbase, self.ampbase, self.lacunarity, self.gain, self.octaves,
            self.base, self.mult,
            self.tex.tuUniform, self.tex.sizeUniform, self.tex.texelUniform
            ]))
        self.tex.tuUniform.assignName('tperms', shader)
        self.tex.sizeUniform.assignName('tpermsSize', shader)
        self.tex.texelUniform.assignName('tpermsTexel', shader)
        
    def randomise(self):
        permst = [x for x in xrange(255)]
        random.shuffle(permst)
        permsts = (''.join((chr(x) + '\xff\xff\xff' for x in permst)))
        permsts += permsts[0:4]
        self.tex.upload(0, 256, permsts)
        
    def uploadData(self, freqbase, ampbase, gain, lacunarity, octaves, base, mult):
        self.freqbase.uploadData(freqbase)
        self.ampbase.uploadData(ampbase)
        self.gain.uploadData(gain)
        self.lacunarity.uploadData(lacunarity)
        self.octaves.uploadData(octaves)
        self.base.uploadData(base)
        self.mult.uploadData(mult)
        

class SimplexProcess(Process):
    
    def render(self, size, seed=None, freqbase=1.0, ampbase=1.0, gain=0.65, 
               lacunarity = 2.0, octaves = 16, range=(0.0, 1.0)):
        if seed is not None: 
            random.seed(seed)
            self.renderer.randomise()
        base = range[0]
        mult = range[1] - range[0]
        self.renderer.uploadData(freqbase, ampbase, gain, lacunarity, octaves, base, mult)
        
        target = TwoDTextureTarget(size, compute = True)
        triangles = self.triangleClass(((0, 0), (0, size), (size, 0), 
                        (size, 0), (size, size), (0, size)), target=target)
        GLCommon.doTasks(target)

        return target.texture

#CITATITON:
#This constant is from Brian Sharpes explanation
#http://briansharpe.wordpress.com/2012/01/13/simplex-noise/#comment-36
#I've independently examined it and think the reasoning is sound. 
#One flaw however is that is is extremely unlikely that for any given
#hashing function that the absolute maxima and minima can be reached, which
#means in practice one can expect the values of the noise function to be
#occupying the middle 80% of possible values rather than the full range.
#Unfortunately, there does not appear to be an easy method to determine actual
#Maxima/Minima short of doing some search based approach. - David Griffin
simplexLen =  (math.sqrt(0.5) / math.sqrt(0.75)) * 0.5
normalFactor = 1.0 / (simplexLen * ((0.5 - simplexLen * simplexLen)**4) * 2.0)

fragmentProg = """
uniform sampler1D tperms;
uniform float tpermsTexel;
uniform float tpermsSize;
uniform float freqBase;
uniform float lacunarity;
uniform float ampBase;
uniform float gain;
uniform int octaves;
uniform float base;
uniform float mult;
const float twopi = """ + str(2 * math.pi) + """;
const float normalFactor = """ + str(normalFactor) + """;
varying vec2 texpos;
const float general_skew = """ + str((math.sqrt(3.0)-1.0)*0.5) + """;
const float general_unskew = """ + str((3.0-math.sqrt(3.0))/6.0) + """;
const float top_general_unskew = - 1.0 + general_unskew + general_unskew;
float noiseFunc(in vec2 pos, in float freq, in float amp);
float texelLookup(in float pos, in float adj);
vec2 gradLookup(in vec2 pos);
float noise(in vec2 grad, in vec2 dis);

float noise(in vec2 grad, in vec2 dis){
    float tempdis = 0.5 - (dis.x * dis.x) - (dis.y * dis.y);
    tempdis = (tempdis > 0.0) ? tempdis : 0.0;
    tempdis *= tempdis;
    tempdis *= tempdis;
    tempdis *= dot(grad, dis);
    return tempdis;
}

float texelLookup(in float pos, in float adj){
    return fract(texture1D(tperms, floor(pos) * tpermsTexel + adj).r);
}

vec2 gradLookup(in vec2 pos){
    float multiplier = texelLookup(pos.x, texelLookup(pos.y, 0.0));
    return vec2(cos(twopi * multiplier), sin(twopi * multiplier));
}

float simplexNoise(in vec2 pos){
    //Return a noise value that is normalised to 0.0 - 1.0;
    //However, whilse values outside of 0.1-0.9 are theoretically 
    //possible, they are exceptionally unlikely.
    float skew_value = (pos.x + pos.y) * general_skew;
    vec2 cornerb = floor(pos + skew_value);
    float unskew_value = (cornerb.x + cornerb.y) * general_unskew;
    vec2 disb = pos - cornerb + unskew_value;
    vec2 cornerm = cornerb + ((disb.x > disb.y) ? vec2(1.0, 0.0) : vec2(0.0, 1.0));
    vec2 cornert = cornerb + 1.0;
    vec2 dism = disb - cornerm + cornerb + general_unskew;
    vec2 dist = disb + top_general_unskew;
    vec3 noise = vec3(noise(gradLookup(cornerb), disb), 
                      noise(gradLookup(cornerm), dism),
                      noise(gradLookup(cornert), dist));
    float ret = (noise.x + noise.y + noise.z) * normalFactor;  
    ret += 1.0;
    ret /= 2.0;
    return ret;
}

void main(){
    float shade = 0.0;
    float frequency = freqBase;
    float amplitude = ampBase;
    float nextAmp = amplitude * gain;
    vec2 texposm = texpos + vec2(101.0, 204.0);
    for (int x = 0; x < octaves - 1; x += 1) {
        //amplitude - nextAmp = the amount that this octave will consume
        //do the last octave seperately to avoid any missing area.
        shade += simplexNoise(texposm * frequency) * (amplitude - nextAmp);
        frequency *= lacunarity;
        amplitude = nextAmp;
        nextAmp *= gain;
    }
    //Note: This is an approximation, and can result in odd looking noise
    //for small numbers of octaves (specifically when gain ** octaves is big,
    //but with an important exception for if octaves = 1 (where results are 
    //fine)
    shade += simplexNoise(texposm * frequency) * amplitude;
    shade *= mult;
    shade += base;
    vec4 outShade = vec4(shade, shade, shade, 1.0);
    gl_FragColor = outShade;
}"""

simplexNoise2d = makeDumbTriangles('SimplexNoise', fragmentProg, 
                 SimplexRenderer, SimplexProcess)
