"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.sprites.image

Provides the classes for hardware accelerated images
"""

from __future__ import division
from collections import namedtuple
from pygl3display.sprites.spritecommon import SpriteCommon, GL3Rect

FloatCoords = namedtuple('FloatCoords', ('left', 'top', 'width', 'height'))

class Image(object):
    """Images are internal things, mainly. Image handles the storing
    of a texture in GPU memory.
    """
    def __init__(self, atlas, intCoords, floatCoords, wrapped):
        """Initialise the image
        Args:
          atlas: the texture atlas the image is on
          intCoords: integer coordinates of the region on the texture atlas
          floatCoords: float coordinates of the region on the texture atlas
          wrapped: If the image is wrapped or not"""
          
        self.atlas = atlas
        self.intCoords = intCoords
        self.coords = floatCoords
        self.rawImage = None
        self.wrapped = wrapped
        
    def getSize(self):
        """Returns the size of the image, not including wrapped area"""
        if self.rawImage is not None:
            return self.rawImage.getSize()
        else:
            return (0, 0)
        
    def reupload(self):
        """Reuploads the image to the texture atlas"""
        if self.rawImage: 
            self.atlas.uploadToBox(self.rawImage, self.intCoords, 
                                   self.wrapped)
        
    def upload(self, image):
        """Upload an image to the atlas
        Arguments:
          image: a RawImage to upload"""
        self.rawImage = image
        self.atlas.uploadToBox(image, self.intCoords, self.wrapped)
        
    def subsurface(self, rect):
        """Return a subsurface of the image
           Arguments:
             rect: a Rect indicating the area to become the subsurface
        """
        # TODO: Test new code
        dx, dy = rect.x, rect.y
        rx, ry = self.intCoords.x + dx, self.intCoords.y + dy
        rwidth = min(dx + rect.width, self.intCoords.width) - dx
        rheight = min(dy + rect.height, self.intCoords.height) - dy
        childIntCoords = GL3Rect(rx, ry, rwidth, rheight, rwidth * rheight)
        
        atlasWidth, atlasHeight = self.atlas.size
        top = childIntCoords.y / atlasHeight
        bottom = (childIntCoords.y + childIntCoords.height) / atlasHeight
        left = childIntCoords.x / atlasWidth
        right = (childIntCoords.x + childIntCoords.width) / atlasWidth
        
        floatCoords = FloatCoords(
                      tl = (left, top),
                      tr = (right, top),
                      br = (right, bottom),
                      bl = (left, bottom)
                      )
        
        return Image(self.atlas, childIntCoords, floatCoords)
        
class GPUImage(Image):
    """A GPUImage is an image which only exists on the GPU, and as such
    has no corresponding RawImage"""
    def getSize(self):
        return self.intCoords.width, self.intCoords.height

class Surface(Image):
    """A Image which is intended to be created by the user. Surfaces are HW
    accelerated."""
    def __init__(self, widthHeight, shader=None, wrapped = False):
        """Initialise the Gl3Surface. 
        Arguments:
          widthHeight: A tuple containing the width / height of the surface
          shader: The shader to use; defaults to shared sprite shader
          wrapped: If the sprite should be wrapped; this results in a different
                   image being uploaded to the shader than if it isn't.
        """
        width, height = widthHeight
        self.wrapped = wrapped
        manager = SpriteCommon.defaultShader if shader is None else shader
        manager.getBox(width, height, self, wrapped)
        self.reuploadData = None
