"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.sprites.spritebuffer

Provides the sprite buffer class, which allows Sprites to be "recycled",
which can provide a modest performance benefit of creating new sprites
"""

from pygl3display.sprites.sprite import Sprite

class SpriteBuffer(object):
    """A buffer for sprites. Creating and deleting sprites is quite taxing
    on the CPU. If a Sprite is allocated a SpriteBuffer, then instead of
    being deleted it may be reset and placed in the buffer for later user.
    The buffer creates sprites as necessary, but will not let sprites which
    are still useful go to waste. To use the sprite buffer, just create one
    and use the createSprite method to create sprites. Then use setX on the
    returned sprite to control the Sprite.
    """
    def __init__(self, length=100, spriteClass=Sprite, 
                 resetFunc = Sprite.reset):
        """Initialise the Sprite Buffer.
         Arguments:
          length: The max number of sprites to keep in the buffer
          spriteClass: the sprites class
          resetFunc: The function to call which resets the sprite to its
                     default state."""
        self.spriteBuffer = []
        self.length = length
        self.spriteClass = spriteClass
        self.resetFunc = resetFunc

    def createSprite(self, *args, **kwargs):
        """Create or get a sprite which is associated with this buffer"""
        kwargs['spriteBuffer'] = self
        return self.spriteBuffer.pop() if self.spriteBuffer else \
               self.spriteClass(*args, **kwargs)
        
    def releaseSprite(self, sprite):
        """Called when a sprite is deleted. If the buffer isn't full
        the sprite is kept. Otherwise, the sprite buffer tells the sprite
        to fully kill itself.
        Arguments:
         sprite: the sprite to potentially save for later"""
        if len(self.spriteBuffer) < self.length: 
            self.spriteBuffer.append(sprite)
            self.resetFunc(sprite)
            return False
        else:
            return True
