"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.sprites.textureatlas

Provides texture atlases
"""
from __future__ import division

import weakref
from pygl3display.infrastructure import GL3Client
from pygl3display.uniforms import UniformStore
from pygl3display.twodtexture import TwoDTexture
from pygl3display.gl import GL_TRIANGLES
from pygl3display.twodshaders import TwoDShaderCommon
from pygl3display.sprites.image import FloatCoords, Image, GPUImage
from pygl3display.sprites.spritecommon import SpriteCommon, GL3Rect, GL3RectNA
from pygl3display.sprites.shader import SpriteShader

__all__ = ['SpriteCommon', 'SpriteShader', 'TextureAtlas2dshader',
           'TextureAtlas2d', 'Image', 'linearFiltering', 'noFiltering']


class CannotBlitIntoError(Exception):
    """Exception raised when a rectangle is larger than the area that it is
    attempting to upload into."""


class TextureAtlas2d(GL3Client):
    """A 2d texture atlas. These are responsible for storing images, and for
    drawing to screen any sprite using an image they contain. Configured for
    optimal use when rendering squares; triangles can be rendered as well, but
    they'll be overallocated slots in the texture atlas."""
    def __init__(self, size, shader = None, wasteWidth = 10, border = 1, 
                 texture = None, rects = None):
        """Initialise the texture array
        Arguments:
         size          : the size of the texture atlas to create
         shader        : the shader which created this atlas
         wasteWidth    : the wasted width tolerable
         border        : the border to give sprites
         texture       : specifies a texture for the atlas to use
         rects         : specifies a list of rectangles that are unallocated
        """
        if shader is None:
            shader = SpriteCommon.defaultShader
            
        super(TextureAtlas2d, self).__init__(GL_TRIANGLES, shader,
                            vertexArraySpec=SpriteCommon.spec)
        
        self.border = border
        self.wasteWidth = wasteWidth
        
        if texture is None:
            self.texture = TwoDTexture(size)
        else:
            self.texture = texture
            
        self.size = self.texture.size # Useful to guarantee consistency 
                                      # when debugging
                                      
        self.addRenderHook(self.texture.bindTexture)
        
        if rects is None:
            self.rects = [GL3Rect(0, 0, self.size, self.size, 
                          self.size*self.size)]
        else:
            self.rects = rects
            
        self.imagerefs = {}
        
        self.uniforms = UniformStore(set([
            TwoDShaderCommon.getProjMatrix(),
            TwoDShaderCommon.defaultCameras.cameraUniform,
            self.texture.tuUniform, self.texture.sizeUniform,
            self.texture.texelUniform
            ]))
        self.texture.tuUniform.assignName('tex', self.shader)
        self.texture.sizeUniform.assignName('texSize', self.shader)
        self.texture.texelUniform.assignName('texTexel', self.shader)
        
    def destroy(self):
        """Destroy the gl texture"""
        self.texture.destroy()
        
    def recreate(self):
        """Recreates the atlas"""
        self.texture.recreate()
        for ref in self.imagesrefs:
            if ref(): 
                ref().reupload()
   
    def getAtlasImage(self, top=0.0, bottom=1.0, left=0.0, right=1.0):
        """Get an image of the atlas for debug purposes.
        Arguments:
          top, bottom, left, right: Floats between 0 and 1 describing
                                    the rectangle to view. Defaults to
                                    entire atlas.
        """
        floatCoords = FloatCoords(
                            top = top, left = left,
                            width = right - left,
                            height = bottom - top            
                      )
        leftEdge, topEdge = left * self.size, top * self.size
        width, height = (right - left) * self.size, \
                        (bottom - top) * self.size
        redrawData = GL3RectNA(leftEdge, topEdge, width, height, width * height)
        return GPUImage(self, redrawData, floatCoords, False)
            
    def getBox(self, width, height, image=None, wrapped=False):
        """Try to get a Box in this atlas. Returns None if it can't be
        allocated. 
        
        Arguments:
         width: The width to find
         height: The height to find
         image : a Image to use; if None create a new Image
         wrapped: If the image should be wrapped
        """
        
        if wrapped:
            width += 2
            height += 2
        
        findWidth = width + self.border
        findHeight = height + self.border
        findSize = findWidth * findHeight
        
        foundRect = None
        bestWastage = self.size * self.size + 1 
        for rect in self.rects:
            if rect.width > findWidth and rect.height > findHeight:
                thisWastage = rect.area - findSize
                if thisWastage < bestWastage:
                    foundRect = rect
                    bestWastage = thisWastage
        if foundRect is None:
            return None
        else:
            # We found a rect to upload to, so split it appropriately
            newX = foundRect.x + findWidth
            remainingX = foundRect.width - findWidth
            newY = foundRect.y + findHeight
            remainingY = foundRect.height - findHeight
            self.rects.remove(foundRect)
            ## Keep as large a box available as possible
            bigBox1 = GL3Rect(newX, foundRect.y, remainingX, foundRect.height,
                              remainingX * foundRect.height)
            bigBox2 = GL3Rect(foundRect.x, newY, foundRect.width, remainingY, 
                              foundRect.width * remainingY)
            if bigBox1.area > bigBox2.area:
                bigBox = bigBox1
                smallBox = GL3Rect(foundRect.x, newY, findWidth, remainingY, 
                                   findWidth * remainingY)
            else:
                bigBox = bigBox2
                smallBox = GL3Rect(newX, foundRect.y, remainingX, findHeight, 
                                   remainingX * findHeight)
            
            if bigBox.area > 0: 
                self.rects.append(bigBox)
            if smallBox.area > 0: 
                self.rects.append(smallBox)
            redrawData = GL3Rect(foundRect.x, foundRect.y, findWidth, 
                                 findHeight, findWidth * findHeight)
            topInt, leftInt, rightInt, bottomInt = \
                    foundRect.y, foundRect.x, newX-1, newY-1
                    
            if wrapped: 
                topInt += 1
                leftInt += 1
                rightInt -= 1
                bottomInt -= 1
                
            top = topInt / self.size
            left = leftInt / self.size
            right = rightInt / self.size
            bottom = bottomInt / self.size
                
            floatCoords = FloatCoords(
                            top = top, left = left,
                            width = right - left,
                            height = bottom - top            
                        )
                        
        if not image:
            ret = Image(self, redrawData, floatCoords, wrapped)
        else:
            Image.__init__(image, self, redrawData, floatCoords, wrapped)
            ret = image
        imgref = weakref.ref(ret, self.deallocateImage)
        self.imagerefs[imgref] = redrawData
        return ret
            
    def deallocateImage(self, imgref):
        """Deallocates an image. This returns the box of the image
           to the pool. It does not defragment so fragmentation will
           happen."""
        if imgref in self.imagerefs:
            self.rects.append(self.imagerefs[imgref])
            del self.imagerefs[imgref]

    def uploadToBox(self, image, rect, wrapped = False):
        """Put an image into a rectangle in the atlas
        Arguments:
         image: an image to upload; must be recognised by the decoder
         rect:  the rect inside the atlas to upload to
         wrapped: Should be true if the image is going to be wrapped
        """
        x, y = rect.x, rect.y
        if wrapped:
            size = image.getSizeWrapped()
            data = image.getRawWrapped()
        else:
            size = image.getSize()
            data = image.getRaw()
        if size > (rect.width, rect.height):
            raise CannotBlitIntoError('Image is bigger than allocated area.')
        self.texture.upload((x, y), size, data)
                        
    def getDefaultFilter(self):
        """Returns the default filter of the shader"""
        return self.shader.defaultFilter
                        
SpriteCommon.setDefaultAtlasClass(TextureAtlas2d)
